from dataclasses import dataclass
from pathlib import Path
from typing import Optional

import pytest

from curated_transformers.tokenizers import Tokenizer
from curated_transformers.tokenizers.chunks import InputChunks, TextChunk

from ..compat import has_hf_transformers, transformers
from ..utils import torch_assertclose
from .util import compare_tokenizer_outputs_with_hf_tokenizer


@dataclass
class _TestModel:
    model: str
    pad_token: Optional[str] = None
    eos_piece: Optional[str] = None


_MODELS = [
    _TestModel("bert-base-cased"),
    _TestModel("camembert-base"),
    _TestModel("EleutherAI/gpt-neox-20b", pad_token="[PAD]", eos_piece="<|endoftext|>"),
    _TestModel("ausboss/llama-30b-supercot", pad_token="</s>", eos_piece="</s>"),
    _TestModel("roberta-base"),
    _TestModel(
        "tiiuae/falcon-7b", pad_token="<|endoftext|>", eos_piece="<|endoftext|>"
    ),
    _TestModel("xlm-roberta-base"),
]

_FRENCH_MODELS = [_TestModel("camembert-base")]


@pytest.fixture
def toy_tokenizer_path():
    return Path(__file__).parent / "toy-roberta"


@pytest.fixture
def toy_tokenizer(toy_tokenizer_path):
    return Tokenizer.from_dir(toy_tokenizer_path)


@pytest.mark.skipif(not has_hf_transformers, reason="requires huggingface transformers")
@pytest.mark.parametrize("model", _MODELS)
def test_against_hf_tokenizers(model, sample_texts):
    compare_tokenizer_outputs_with_hf_tokenizer(
        sample_texts, model.model, Tokenizer, pad_token=model.pad_token
    )


@pytest.mark.skipif(not has_hf_transformers, reason="requires huggingface transformers")
@pytest.mark.parametrize("model", _MODELS)
def test_against_hf_tokenizers_short(model, short_sample_texts):
    compare_tokenizer_outputs_with_hf_tokenizer(
        short_sample_texts, model.model, Tokenizer, pad_token=model.pad_token
    )


@pytest.mark.skipif(not has_hf_transformers, reason="requires huggingface transformers")
@pytest.mark.parametrize("model", _FRENCH_MODELS)
def test_against_hf_tokenizers_french(model, french_sample_texts):
    compare_tokenizer_outputs_with_hf_tokenizer(
        french_sample_texts, model.model, Tokenizer, pad_token=model.pad_token
    )


@pytest.mark.parametrize("model", _MODELS)
def test_special_pieces(model):
    tokenizer = Tokenizer.from_hf_hub(name=model.model)
    assert tokenizer.eos_piece == model.eos_piece


@pytest.mark.skipif(not has_hf_transformers, reason="requires huggingface transformers")
def test_from_dir(toy_tokenizer, toy_tokenizer_path, sample_texts):
    hf_tokenizer = transformers.RobertaTokenizerFast.from_pretrained(
        str(toy_tokenizer_path)
    )

    pieces = toy_tokenizer(sample_texts)
    hf_pieces = hf_tokenizer(sample_texts, padding=True, return_tensors="pt")
    torch_assertclose(
        pieces.padded_tensor(padding_id=hf_tokenizer.pad_token_id),
        hf_pieces["input_ids"].int(),
    )


@pytest.mark.skipif(not has_hf_transformers, reason="requires huggingface transformers")
def test_from_json(toy_tokenizer_path, sample_texts):
    with open(toy_tokenizer_path / "tokenizer.json", encoding="utf-8") as f:
        tokenizer = Tokenizer.from_json(f.read())
    hf_tokenizer = transformers.RobertaTokenizerFast.from_pretrained(
        str(toy_tokenizer_path)
    )

    pieces = tokenizer(sample_texts)
    hf_pieces = hf_tokenizer(sample_texts, padding=True, return_tensors="pt")
    torch_assertclose(
        pieces.padded_tensor(padding_id=hf_tokenizer.pad_token_id),
        hf_pieces["input_ids"].int(),
    )


def test_invalid_string_input(toy_tokenizer):
    with pytest.raises(ValueError, match=r"Non-string.*float, int"):
        toy_tokenizer(["hello", 42, 3.14159])


def test_invalid_chunk_input(toy_tokenizer):
    with pytest.raises(ValueError, match=r"Non-chunk.*float, int"):
        toy_tokenizer([InputChunks([TextChunk("hello")]), 42, 3.14159])



from abc import ABC, abstractmethod
from collections import UserList
from typing import Iterable

import torch
from torch import Tensor


class LogitsTransform(ABC):
    """
    A logits transform changes the logits of a softmax distribution in some
    way. For instance, :class:`TemperatureTransform` changes the temperature
    of the softmax distribution.
    """

    def __call__(self, logits: Tensor, inplace: bool = False) -> Tensor:
        """
        Transform the given logits.

        :param logits:
            The logits array.

            *Shape:* ``(..., n_class)``
        :param inplace:
            Transform logits in-place.
        :returns:
            The transformed logits. This will be the same tensor object as
            the ``logits`` argument when ``inplace`` is set.

            *Shape:* ``(..., n_class)``
        """
        if not inplace:
            logits = logits.clone()
        self._process_logits(logits)
        return logits

    @abstractmethod
    def _process_logits(self, logits: Tensor):
        """
        Transform logits in-place.

        :param logits:
            The logits array.

            *Shape:* ``(..., n_class)``
        """
        ...


class CompoundLogitsTransform(UserList, LogitsTransform):
    """
    Sequentially apply multiple logit transforms.
    """

    def _process_logits(self, logits: Tensor):
        for transform in self:
            logits = transform(logits, inplace=True)


class TopKTransform(LogitsTransform):
    """
    Set the probability of non-top-k classes to zero. The probability of
    the classes that are zeroed out is redistributed across the top-k
    classes.
    """

    def __init__(self, k: int):
        """
        Construct a top-k logits transform.

        :param k:
            The value of k in top-k. The transform is a no-op for values
            less than 1.
        """
        super().__init__()
        self.k = k

    def _process_logits(self, logits: Tensor):
        if self.k < 1:
            return

        # Nothing to do if we have k or fewer classes.
        if logits.size(-1) <= self.k:
            return

        mask = torch.finfo(logits.dtype).min
        cutoff = torch.topk(logits, self.k, dim=-1).values[..., -1:]
        logits[logits < cutoff] = mask


class TopPTransform(LogitsTransform):
    """
    Keep the smallest possible set of most probable vocab items, such that
    their cumulative probability is >= p. Sampling using the top-p transform
    is also known as nucleus sampling (`Holzman et al., 2019`_). The
    probability of the items that are masked out is redistributed across the
    top-p items.

    .. _Holzman et al., 2019: https://arxiv.org/abs/1904.09751
    """

    def __init__(self, p: float):
        """
        Construct a top-p logits transform.

        :param p:
            The value of p in top-p. The transform is a no-op for
            ``p = 1.0``.
        """
        super().__init__()
        if not 0.0 <= p <= 1.0:
            raise ValueError(
                f"Top-p probability must be between 0 and 1 (inclusive), was: {p}"
            )
        self.p = p

    def _process_logits(self, logits: Tensor):
        if self.p == 1.0:
            return

        sorted_values, sorted_indices = torch.sort(logits, descending=True)
        cumulative_probs = sorted_values.softmax(-1).cumsum(-1)

        retain_indices = cumulative_probs < self.p

        # We need to get the smallest set for which the cumulative
        # probability exceeds p, so we need an additional logit. We
        # never include one logit too many because of the use of
        # ``<`` rather than `<=` above.
        retain_indices = retain_indices.roll(1, dims=-1)
        retain_indices[:, 0] = True

        # Sorted values with masked-out logits.
        mask = torch.finfo(logits.dtype).min
        sorted_values = torch.where(retain_indices, sorted_values, mask)

        # Put back in the unsorted logits.
        logits.scatter_(-1, sorted_indices, sorted_values)


class TemperatureTransform(LogitsTransform):
    """
    Apply temperature to the softmax distribution. Given the temperature ``T``
    and logits ``z(y|x)``:

    .. math::
        p(y|x) = softmax(z(y|x)/T)

    For a temperature ``T``:

    - ``T = 1``: the distribution is not changed.
    - ``T < 1``: the entropy of the distribution is decreased.
    - ``T > 1``: the entropy of the distribution is increased.
    """

    def __init__(self, temperature: float = 1.0):
        """
        Create a temperature transform with a given temperature.

        :param temperature:
            The temperature. Must be a non-zero positive value.
        """
        super().__init__()
        if temperature <= 0.0:
            raise ValueError(
                f"The temperature must be a non-zero positive value, was: {temperature}"
            )
        self.temperature = temperature

    def _process_logits(self, logits: Tensor):
        if self.temperature == 1.0:
            return

        logits /= self.temperature


class VocabMaskTransform(LogitsTransform):
    """
    Set the probability of specific vocabulary pieces to zero.
    """

    def __init__(self, pieces_to_mask: Iterable[int]):
        """
        Construct a mask logits transform.

        :param pieces_to_mask:
            Identifers pertaining to the vocabulary pieces that
            need to be masked. An empty iterable results in a no-op.
        """
        super().__init__()

        self.pieces_to_mask = torch.tensor(
            list(pieces_to_mask), dtype=torch.long
        )  # `torch.int32` isn't supported in indexing operations prior in torch<2.0.0.

        if self.pieces_to_mask.dim() != 1:
            raise ValueError("Vocabulary piece mask must be 1D")
        elif (
            self.pieces_to_mask.size(dim=-1) != 0
            and int(self.pieces_to_mask.min(dim=-1).values) < 0
        ):
            raise ValueError("Vocabulary piece identifiers must be >= 0")

    def _process_logits(self, logits: Tensor):
        if self.pieces_to_mask.size(dim=-1) == 0:
            return

        try:
            mask = torch.finfo(logits.dtype).min
            logits[..., self.pieces_to_mask] = mask
        except IndexError:
            max_expected = logits.size(dim=-1)
            max_received = int(self.pieces_to_mask.max(dim=-1).values)
            if max_received >= max_expected:
                raise ValueError(
                    f"Vocabulary piece identifiers must be < {max_expected}, but got {max_received}"
                )



import os
from abc import ABC, abstractmethod
from typing import IO, Optional


class RepositoryFile(ABC):
    """
    A repository file.

    Repository files can be a local path or a remote path exposed as a
    file-like object. This is a common base class for such different types
    of repository files.
    """

    @abstractmethod
    def open(self, mode: str = "rb", encoding: Optional[str] = None) -> IO:
        """
        Get the file as a file-like object.

        :param mode:
            Mode to open the file with (see Python ``open``).
        :param encoding:
            Encoding to use when the file is opened as text.
        :returns:
            An I/O stream.
        :raises FileNotFoundError:
            When the file cannot be found.
        :raises OSError:
            When the file cannot be opened.
        """
        ...

    @property
    @abstractmethod
    def path(self) -> Optional[str]:
        """
        Get the file as a local path.

        :returns:
            The repository file. If the file is not available as a local
            path, the value of this property is ``None``. In these cases
            ``open`` can be used to get the file as a file-like object.
        """
        ...

    @abstractmethod
    def exists(self) -> bool:
        """
        Returns if the file exists. This can cause the file
        to be cached locally.
        """
        ...


class LocalFile(RepositoryFile):
    """
    Repository file on the local machine.
    """

    def __init__(self, path: str):
        """
        Construct a local file representation.

        :param path:
            The path of the file on the local filesystem.
        """
        super().__init__()
        self._path = path

    def open(self, mode: str = "rb", encoding: Optional[str] = None) -> IO:
        # Raises OSError, so we don't have to do any rewrapping.
        return open(self._path, mode=mode, encoding=encoding)

    @property
    def path(self) -> Optional[str]:
        return self._path

    def exists(self) -> bool:
        return os.path.isfile(self._path)



from dataclasses import dataclass
from typing import Generic, List, Optional, TypeVar

from torch import Tensor

from ..layers.cache import CacheProtocol

CacheT = TypeVar("CacheT", bound=CacheProtocol)


@dataclass
class ModelOutput:
    """
    Base class for model outputs.

    :param all_outputs:
        The first element is the output of the embedding layer. The
        rest of the elements are the states of each encoder hidden
        layer respectively.
    """

    all_outputs: List[Tensor]

    @property
    def embedding_layer(self) -> Tensor:
        """
        Return the output of the embedding layer.

        :returns:
            Embedding layer output.

            *Shape:* ``(batch_size, seq_len, width)``
        """
        return self.all_outputs[0]

    def hidden_layer_states(self, idx: int) -> Tensor:
        """
        Return the hidden representations of a given layer.

        :param idx:
            Layer index. Must be in ``[0, n_hidden_layers)``.
        :returns:
            Hidden representation of the layer.

            *Shape:* ``(batch_size, seq_len, width)``
        """
        if 0 <= idx < len(self.all_outputs) - 1:
            return self.all_outputs[idx + 1]
        else:
            raise ValueError(
                "Attempting to select a transformer output tensor using an invalid "
                f"layer index ({idx}). Expected range: 0 <= idx < {(len(self.all_outputs) - 1)}"
            )

    @property
    def last_hidden_layer_state(self) -> Tensor:
        """
        Return the hidden representation of the last layer.

        :returns:
            Last hidden representation of the last layer.

            *Shape:* ``(batch_size, seq_len, width)``
        """
        return self.all_outputs[-1]

    @property
    def all_hidden_layer_states(self) -> List[Tensor]:
        """
        Return the hidden representation of all the layers.

        :returns:
            Hidden representations of all the layers.

            *Shape:* ``(batch_size, seq_len, width)``
        """
        return self.all_outputs[1:]


@dataclass
class ModelOutputWithCache(Generic[CacheT], ModelOutput):
    """
    Output of decoder modules.

    :param cache:
        Model cache. The cache can be used with future calls
        to a model to reuse computations for efficiency
    """

    cache: Optional[List[CacheT]]


@dataclass
class CausalLMOutputWithCache(Generic[CacheT], ModelOutputWithCache[CacheT]):
    """
    Output of causal language model modules.

    :param logits:
        Logits of the distributions of predicted tokens.
    """

    logits: Tensor



from functools import partial
from typing import Any, Dict, Mapping, Optional, Tuple, Type, TypeVar

import torch
from torch import Tensor
from torch.nn import Dropout, LayerNorm

from ...layers.attention import (
    AttentionHeads,
    QkvMode,
    QkvSplitGroupedByKVHeads,
    ScaledDotProductAttention,
    SelfAttention,
)
from ...layers.feedforward import PointwiseFeedForward
from ...layers.transformer import (
    EmbeddingDropouts,
    EmbeddingLayerNorms,
    EncoderLayer,
    TransformerDropouts,
    TransformerLayerNorms,
)
from ..hf_hub import FromHF
from ..hf_hub.conversion import state_dict_from_hf, state_dict_to_hf
from ..transformer import TransformerEncoder
from ._hf import HF_PARAM_KEY_TRANSFORMS, _config_from_hf, _config_to_hf
from .config import RoBERTaConfig
from .embeddings import RoBERTaEmbeddings

# Only provided as typing.Self in Python 3.11+.
Self = TypeVar("Self", bound="RoBERTaEncoder")


class RoBERTaEncoder(TransformerEncoder[RoBERTaConfig], FromHF[RoBERTaConfig]):
    """
    RoBERTa (`Liu et al., 2019`_) encoder.

    .. _Liu et al., 2019: https://arxiv.org/abs/1907.11692
    """

    def __init__(self, config: RoBERTaConfig, *, device: Optional[torch.device] = None):
        """
        Construct a RoBERTa encoder.

        :param config:
            Encoder configuration.
        :param device:
            Device to which the module is to be moved.
        :returns:
            The encoder.
        """
        super().__init__(config)

        self.embeddings = RoBERTaEmbeddings(
            dropouts=EmbeddingDropouts(
                embed_output_dropout=Dropout(config.embedding.dropout_prob)
            ),
            embedding_width=config.embedding.embedding_width,
            hidden_width=config.layer.feedforward.hidden_width,
            layer_norms=EmbeddingLayerNorms(
                embed_output_layer_norm=LayerNorm(
                    config.embedding.embedding_width, config.embedding.layer_norm_eps
                )
            ),
            n_pieces=config.embedding.n_pieces,
            n_positions=config.embedding.n_positions,
            n_types=config.embedding.n_types,
            padding_id=config.padding_id,
        )
        self.max_seq_len = config.model_max_length

        hidden_width = config.layer.feedforward.hidden_width
        layer_norm = partial(
            LayerNorm,
            hidden_width,
            config.layer.layer_norm_eps,
            device=device,
        )
        self.layers = torch.nn.ModuleList(
            [
                EncoderLayer(
                    attention_layer=SelfAttention(
                        attention_heads=AttentionHeads.uniform(
                            config.layer.attention.n_query_heads,
                            QkvSplitGroupedByKVHeads(),
                        ),
                        attention_scorer=ScaledDotProductAttention(
                            dropout_prob=config.layer.attention.dropout_prob,
                            linear_biases=None,
                        ),
                        hidden_width=hidden_width,
                        qkv_mode=QkvMode.SEPARATE,
                        rotary_embeds=None,
                        use_bias=config.layer.attention.use_bias,
                        device=device,
                    ),
                    feed_forward_layer=PointwiseFeedForward(
                        activation=config.layer.feedforward.activation.module(),
                        hidden_width=hidden_width,
                        intermediate_width=config.layer.feedforward.intermediate_width,
                        use_bias=config.layer.feedforward.use_bias,
                        use_gate=config.layer.feedforward.use_gate,
                        device=device,
                    ),
                    dropouts=TransformerDropouts.layer_output_dropouts(
                        config.layer.dropout_prob
                    ),
                    layer_norms=TransformerLayerNorms(
                        attn_residual_layer_norm=layer_norm(),
                        ffn_residual_layer_norm=layer_norm(),
                    ),
                    use_parallel_attention=config.layer.attention.use_parallel_attention,
                )
                for _ in range(config.layer.n_hidden_layers)
            ]
        )

    @classmethod
    def is_supported(cls: Type[Self], config: Dict[str, Any]) -> bool:
        return config.get("model_type") == "roberta"

    @classmethod
    def state_dict_from_hf(
        cls: Type[Self], params: Mapping[str, Tensor]
    ) -> Mapping[str, Tensor]:
        return state_dict_from_hf(params, HF_PARAM_KEY_TRANSFORMS)

    @classmethod
    def state_dict_to_hf(
        cls: Type[Self], params: Mapping[str, Tensor]
    ) -> Mapping[str, Tensor]:
        return state_dict_to_hf(params, HF_PARAM_KEY_TRANSFORMS)

    @classmethod
    def config_from_hf(cls, hf_config: Mapping[str, Any]) -> RoBERTaConfig:
        return _config_from_hf(hf_config)

    @classmethod
    def config_to_hf(cls, curated_config: RoBERTaConfig) -> Mapping[str, Any]:
        return _config_to_hf(curated_config)

    @classmethod
    def from_hf_config(
        cls: Type[Self],
        *,
        hf_config: Any,
        device: Optional[torch.device] = None,
    ) -> Self:
        config = cls.config_from_hf(hf_config)
        return cls(config, device=device)



import pytest

from curated_transformers.models.electra.encoder import ELECTRAEncoder

from ...compat import has_hf_transformers
from ...conftest import TORCH_DEVICES
from ..util import assert_encoder_output_equals_hf


@pytest.mark.skipif(not has_hf_transformers, reason="requires huggingface transformers")
@pytest.mark.parametrize("torch_device", TORCH_DEVICES)
@pytest.mark.parametrize("with_torch_sdp", [False, True])
@pytest.mark.parametrize(
    "model_name",
    ["explosion-testing/electra-test"],
)
def test_encoder(model_name: str, torch_device, with_torch_sdp):
    assert_encoder_output_equals_hf(
        ELECTRAEncoder,
        model_name,
        torch_device,
        with_torch_sdp=with_torch_sdp,
    )



from .file import LocalFile, RepositoryFile
from .fsspec import FsspecArgs, FsspecFile, FsspecRepository
from .hf_hub import HfHubFile, HfHubRepository
from .repository import ModelRepository, Repository, TokenizerRepository
from .transaction import TransactionContext

__all__ = [
    "FsspecArgs",
    "FsspecFile",
    "FsspecRepository",
    "HfHubFile",
    "HfHubRepository",
    "LocalFile",
    "ModelRepository",
    "Repository",
    "RepositoryFile",
    "TokenizerRepository",
    "TransactionContext",
]



from typing import Iterable, List, Optional

from curated_tokenizers import SentencePieceProcessor

from ..chunks import MergedInputChunks, MergedSpecialPieceChunk
from ..tokenizer import PiecesWithIds
from .legacy_tokenizer import LegacyTokenizer


class SentencePieceTokenizer(LegacyTokenizer):
    """
    Piece tokenizer using SentencePiece encoding (`Kudo et al., 2018`_).

    .. _Kudo et al., 2018: https://arxiv.org/abs/1808.06226
    """

    def __init__(
        self,
        *,
        processor: SentencePieceProcessor,
    ):
        """
        Construct a tokenizer from ``curated-tokenizers`` SentencePiece processor.

        :param processor:
            The processor to wrap.
        """
        self.processor = processor
        self._eos_piece = processor.id_to_piece(processor.eos_id())

    @property
    def eos_piece(self) -> Optional[str]:
        return self._eos_piece

    def piece_to_id(self, piece: str) -> Optional[int]:
        return self.processor.piece_to_id(piece)

    def _decode(
        self, input: Iterable[Iterable[int]], skip_special_pieces: bool
    ) -> List[str]:
        # skip_special_pieces is currently ignored. Since sentencepiece
        # processes the whole input output, this probably needs to be
        # handled by the sentencepiece library itself?
        return [self.processor.decode_from_ids(ids) for ids in input]

    def _encode(self, input: Iterable[MergedInputChunks]) -> PiecesWithIds:
        ids = []
        pieces = []

        for seq in input:
            seq_ids = []
            seq_pieces = []

            for chunk in seq:
                if isinstance(chunk, MergedSpecialPieceChunk):
                    piece_id = self.processor.piece_to_id(chunk.piece)
                    if piece_id is None:
                        raise ValueError(f"Unknown special piece: {chunk.piece}")
                    seq_ids.append(piece_id)
                    seq_pieces.append(chunk.piece)
                else:
                    chunk_ids, chunk_pieces = self.processor.encode(chunk.text)
                    seq_ids.extend(chunk_ids)
                    seq_pieces.extend(chunk_pieces)

            ids.append(seq_ids)
            pieces.append(seq_pieces)

        return PiecesWithIds(ids=ids, pieces=pieces)



from typing import Any, Dict, List, Mapping, Optional, Tuple

import torch

from ...util.string import StringTransform, StringTransformations
from ..hf_hub.conversion import (
    CommonHFKeys,
    HFConfigKey,
    HFConfigKeyDefault,
    HFSpecificConfig,
    config_from_hf,
    config_to_hf,
)
from .config import BERTConfig

# Order-dependent.
HF_PARAM_KEY_TRANSFORMS: List[StringTransform] = [
    # Old HF parameter names (one-way transforms).
    StringTransformations.regex_sub((r"\.gamma$", ".weight"), backward=None),
    StringTransformations.regex_sub((r"\.beta$", ".bias"), backward=None),
    # Prefixes.
    StringTransformations.remove_prefix("bert.", reversible=False),
    StringTransformations.regex_sub(
        (r"^encoder\.(layer\.)", "\\1"),
        (r"^(layer\.)", "encoder.\\1"),
    ),
    # Layers.
    StringTransformations.regex_sub((r"^layer", "layers"), (r"^layers", "layer")),
    # Attention blocks.
    StringTransformations.regex_sub(
        (r"\.attention\.self\.(query|key|value)", ".mha.\\1"),
        (r"\.mha\.(query|key|value)", ".attention.self.\\1"),
    ),
    StringTransformations.sub(".attention.output.dense", ".mha.output"),
    StringTransformations.sub(
        r".attention.output.LayerNorm", ".attn_residual_layer_norm"
    ),
    # Pointwise feed-forward layers.
    StringTransformations.sub(".intermediate.dense", ".ffn.intermediate"),
    StringTransformations.regex_sub(
        (r"(\.\d+)\.output\.LayerNorm", "\\1.ffn_residual_layer_norm"),
        (r"(\.\d+)\.ffn_residual_layer_norm", "\\1.output.LayerNorm"),
    ),
    StringTransformations.regex_sub(
        (r"(\.\d+)\.output\.dense", "\\1.ffn.output"),
        (r"(\.\d+)\.ffn\.output", "\\1.output.dense"),
    ),
    # Embeddings.
    StringTransformations.replace(
        "embeddings.word_embeddings.weight", "embeddings.piece_embeddings.weight"
    ),
    StringTransformations.replace(
        "embeddings.token_type_embeddings.weight", "embeddings.type_embeddings.weight"
    ),
    StringTransformations.replace(
        "embeddings.position_embeddings.weight", "embeddings.position_embeddings.weight"
    ),
    StringTransformations.replace(
        "embeddings.LayerNorm.weight", "embeddings.embed_output_layer_norm.weight"
    ),
    StringTransformations.replace(
        "embeddings.LayerNorm.bias", "embeddings.embed_output_layer_norm.bias"
    ),
]

HF_CONFIG_KEYS: List[Tuple[HFConfigKey, Optional[HFConfigKeyDefault]]] = [
    (CommonHFKeys.ATTENTION_PROBS_DROPOUT_PROB, None),
    (CommonHFKeys.DTYPE, HFConfigKeyDefault("float32")),
    (CommonHFKeys.HIDDEN_DROPOUT_PROB, None),
    (CommonHFKeys.HIDDEN_SIZE, None),
    (CommonHFKeys.HIDDEN_ACT, None),
    (CommonHFKeys.INTERMEDIATE_SIZE, None),
    (CommonHFKeys.LAYER_NORM_EPS, None),
    (CommonHFKeys.NUM_ATTENTION_HEADS_UNIFORM, None),
    (CommonHFKeys.NUM_HIDDEN_LAYERS, None),
    (CommonHFKeys.VOCAB_SIZE, None),
    (CommonHFKeys.TYPE_VOCAB_SIZE, None),
    (CommonHFKeys.MAX_POSITION_EMBEDDINGS, None),
]

HF_SPECIFIC_CONFIG = HFSpecificConfig(architectures=["BertModel"], model_type="bert")


def _config_from_hf(hf_config: Mapping[str, Any]) -> BERTConfig:
    kwargs = config_from_hf("BERT", hf_config, HF_CONFIG_KEYS)
    return BERTConfig(
        embedding_width=CommonHFKeys.HIDDEN_SIZE.get_kwarg(kwargs),
        model_max_length=CommonHFKeys.MAX_POSITION_EMBEDDINGS.get_kwarg(kwargs),
        **kwargs,
    )


def _config_to_hf(curated_config: BERTConfig) -> Dict[str, Any]:
    out = config_to_hf(curated_config, [k for k, _ in HF_CONFIG_KEYS])
    return HF_SPECIFIC_CONFIG.merge(out)



from typing import Type, TypeVar, Union


class Default:
    """
    Marker type to be used as a default value for
    :py:class:`~curated_transformers.semver.FutureMandatory`.
    """

    def __init__(self):
        raise TypeError(
            f"{type(self).__name__} is a marker class and cannot be constructed."
        )


T = TypeVar("T")

#: ``FutureMandatory[T]`` is either an instance of ``T`` or the type
#: :py:class:`~curated_transformers.semver.Default`. It used as the
#: type for arguments that will become mandatory in the next major
#: version.
FutureMandatory = Union[T, Type[Default]]



from dataclasses import dataclass
from typing import Optional, Tuple

import torch
from torch import Tensor
from torch.nn import Dropout, Embedding, Identity, Linear, Module

from .attention import AttentionMask, KeyValueCache, SelfAttention
from .feedforward import PointwiseFeedForward


@dataclass
class EmbeddingLayerNorms:
    """
    Layer normalizations used in a transformer embedding layer.

    By default, all the normalizations are disabled by setting the layer
    normalization to the Torch ``Identity`` module. Therefore, only
    normalizations that are needed have to be set.

    :param embed_output_layer_norm:
        Normalization of the embeddings.
    :param proj_output_layer_norm:
        Normalization of the output of the projection layer.
    """

    embed_output_layer_norm: Module = Identity()
    proj_output_layer_norm: Module = Identity()


@dataclass
class EmbeddingDropouts:
    """
    Dropouts used in a transformer embedding layer.

    By default, all the dropouts are disabled by setting the dropout
    to the Torch ``Identity`` module. Therefore, only dropouts that are
    needed have to be set.

    :param embed_output_dropout:
        Dropout of the embeddings.
    :param proj_output_dropout:
        Dropout of the output of the projection layer.
    """

    embed_output_dropout: Module = Identity()
    proj_output_dropout: Module = Identity()


class TransformerEmbeddings(Module):
    """
    Transformer embeddings layer.

    This is a generic transformer embedding layer. The layer always has piece
    embeddings and can optionally have position embeddings, type embeddings,
    and a projection of embeddings to the model's hidden size.
    """

    def __init__(
        self,
        *,
        dropouts: EmbeddingDropouts,
        embedding_width: int,
        hidden_width: int,
        layer_norms: EmbeddingLayerNorms,
        n_pieces: int,
        n_positions: Optional[int],
        n_types: Optional[int],
        device: Optional[torch.device] = None,
    ):
        """
        Construct an embeddings layer.

        :param dropouts:
            Dropouts to use in the embeddings layer.
        :param embedding_width:
            Width of the embeddings.
        :param hidden_width:
            Hidden width of the transformer. If this width differs from
            ``embedding_width``, a projection layer is added to ensure
            that the output of the embeddings layer has the same width as the
            transformer.
        :param layer_norms:
            Layer norms to use in the embeddings layer.
        :param n_pieces:
            Number of piece embeddings.
        :param n_positions:
            Number of position embeddings. Position embeddings are disabled
            by using ``None``. Position embeddings can be used to inform the
            model of input order.
        :param n_types:
            Number of type embeddings. Type embeddings are disabled by using
            ``None``. Type embeddings can be used to inform the model of the
            spans of different sequences in the input.
        :param device:
            Device on which the module is to be initialized.
        """
        super().__init__()

        self.piece_embeddings = Embedding(
            num_embeddings=n_pieces,
            embedding_dim=embedding_width,
            device=device,
        )

        if n_types is None:
            self.type_embeddings = None
        else:
            self.type_embeddings = Embedding(
                num_embeddings=n_types,
                embedding_dim=embedding_width,
                device=device,
            )

        if n_positions is None:
            self.position_embeddings = None
        else:
            self.position_embeddings = Embedding(
                num_embeddings=n_positions,
                embedding_dim=embedding_width,
                device=device,
            )

        if embedding_width != hidden_width:
            self.projection: Linear = Linear(
                embedding_width,
                hidden_width,
                device=device,
            )
        else:
            self.projection = None  # type: ignore

        self.embed_output_layer_norm = layer_norms.embed_output_layer_norm
        self.proj_output_layer_norm = layer_norms.proj_output_layer_norm

        self.embed_output_dropout = dropouts.embed_output_dropout
        self.proj_output_dropout = dropouts.proj_output_dropout

    def _get_positions(self, x: Tensor) -> Tensor:
        return torch.arange(x.shape[1], device=x.device).expand(1, -1)

    def _get_type_ids(self, x: Tensor) -> Tensor:
        return torch.zeros_like(x)

    def forward(
        self,
        piece_ids: Tensor,
        *,
        positions: Optional[Tensor] = None,
        type_ids: Optional[Tensor] = None,
    ) -> Tensor:
        """
        Apply the embedding layer to the piece identifiers.

        :param piece_ids:
            Piece identifiers to embed.

            *Shape:* ``(batch_size, seq_len)``
        :param positions:
            Positional with which to fetch the positional embeddings for the
            sequences.

            *Shape:* ``(batch_size, seq_len)``
        :param type_ids:
            Type identifiers to indicate the spans of different
            sequences in the input. Useful when performing tasks like
            sequence classification and question answering.

            *Shape:* ``(batch_size, seq_len)``
        """

        embeddings = self.piece_embeddings(piece_ids)

        if self.type_embeddings is not None:
            if type_ids is None:
                type_ids = self._get_type_ids(piece_ids)
            embeddings += self.type_embeddings(type_ids)

        if self.position_embeddings is not None:
            if positions is None:
                positions = self._get_positions(piece_ids)
            position_embeddings = self.position_embeddings(positions)
            embeddings += position_embeddings

        embeddings = self.embed_output_layer_norm(embeddings)
        embeddings = self.embed_output_dropout(embeddings)

        if self.projection is not None:
            embeddings = self.projection(embeddings)
            embeddings = self.proj_output_layer_norm(embeddings)
            return self.proj_output_dropout(embeddings)
        else:
            return embeddings


@dataclass
class TransformerLayerNorms:
    """
    Layer normalizations used in a transformer layer.

    By default, all the normalizations are disabled by setting the layer
    normalization to the Torch ``Identity`` module. Therefore, only
    normalizations that are needed have to be set.

    :param attn_input_layer_norm:
        Normalization of the input to the attention layer.
    :param attn_residual_layer_norm:
        Normalization of the output of the attention layer after the
        residual connection.
    :param ffn_input_layer_norm:
        Normalization of the input to the feed-forward layer.
    :param ffn_residual_layer_norm:
        Normalization of the output of the feed-forward layer after the
        residual connection.
    """

    attn_input_layer_norm: Module = Identity()
    attn_residual_layer_norm: Module = Identity()
    ffn_input_layer_norm: Module = Identity()
    ffn_residual_layer_norm: Module = Identity()


@dataclass
class TransformerDropouts:
    """
    Dropouts used in a transformer layer.

    By default, all the dropouts are disabled by setting the dropout
    to the Torch ``Identity`` module. Therefore, only dropouts that are
    needed have to be set.

    :param attn_output_dropout:
        Dropout of the output of the attention layer.
    :param ffn_output_dropout:
        Dropout of the output of the attention layer.
    :param parallel_attn_dropout:
        Dropout after summing the attention and feed-forward layers. Only
        used when parallel attention is enabled.
    """

    attn_output_dropout: Module = Identity()
    ffn_output_dropout: Module = Identity()
    parallel_attn_dropout: Module = Identity()

    @classmethod
    def layer_output_dropouts(cls, p: float) -> "TransformerDropouts":
        """
        Utility method to construct attention and feed-forward layer
        dropouts.

        :param p:
            Dropout probability.
        :returns:
            Dropouts of attention and feed-forward layers set to ``p``.
        """
        return TransformerDropouts(
            attn_output_dropout=Dropout(p), ffn_output_dropout=Dropout(p)
        )

    @classmethod
    def parallel_attention_dropout(cls, p: float) -> "TransformerDropouts":
        """
        Utility method to construct parallel attention dropout.

        :param p:
            Dropout probability.
        :returns:
            Dropouts of parallel attention set to ``p``.
        """
        return TransformerDropouts(parallel_attn_dropout=Dropout(p))


class _TransformerLayer(Module):
    """
    Transformer decoder layer (`Vaswani et al., 2017`_) base class.

    This is a generic transformer layer. :py:class:`DecoderLayer` and
    :py:class:`EncoderLayer` provide specialized encoder/decoder layers.

    .. _Vaswani et al., 2017: https://arxiv.org/abs/1706.03762
    """

    # Don't expose this module, we cannot provide the full forward method
    # without violating the Liskov substitution principle. Ideally, we'd
    # use this module using composition inside the encoder/decoder layers,
    # but we can't without adding a prefix.

    def __init__(
        self,
        *,
        attention_layer: SelfAttention,
        dropouts: TransformerDropouts,
        feed_forward_layer: PointwiseFeedForward,
        layer_norms: TransformerLayerNorms,
        use_parallel_attention: bool,
    ):
        """
        Construct a transformer layer.

        :param attention_layer:
            The attention layer to use in the transformer layer.
        :param dropouts:
            Dropouts to use in the transformer layer.
        :param feed_forward_layer:
            The pointwise feed-forward layer to use in the transformer layer.
        :param layer_norms:
            Layer norms to use in the layer.
        :param use_parallel_attention:
             Use parallel attention.
        """
        super().__init__()

        self.use_parallel_attention = use_parallel_attention

        self.mha = attention_layer

        self.ffn = feed_forward_layer

        self.attn_output_dropout = dropouts.attn_output_dropout
        self.ffn_output_dropout = dropouts.ffn_output_dropout
        self.parallel_attn_dropout = dropouts.parallel_attn_dropout

        self.attn_input_layer_norm = layer_norms.attn_input_layer_norm
        self.attn_residual_layer_norm = layer_norms.attn_residual_layer_norm
        self.ffn_input_layer_norm = layer_norms.ffn_input_layer_norm
        self.ffn_residual_layer_norm = layer_norms.ffn_residual_layer_norm

    def _forward(
        self,
        input: Tensor,
        attention_mask: AttentionMask,
        *,
        cache: Optional[KeyValueCache] = None,
        positions: Optional[Tensor] = None,
        store_cache: bool = False,
        use_causal_mask: bool,
    ) -> Tuple[Tensor, Optional[KeyValueCache]]:
        """
        Apply the transformer layer to the given piece hidden representations.

        :param input:
            Hidden representations to apply the layer to.

            *Shape:* ``(batch_size, seq_len, width)``
        :param attention_mask:
            Attention mask. Sequence elements for which the
            corresponding mask element is set to ``False`` are ignored
            during attention calculation.
        :param use_causal_mask:
            Mask out succeeding sequence elements when ``True``.
        :param cache:
            Key/value cache to avoid recomputing
            key/value representations for tokens that were previously seen.
        :param positions:
            Input positions. Positions are needed to look up rotary embeddings.
            Normally, these positions are calculated automatically. But if the
            positions deviate for some reason, they can be provided through this argument.
        :param store_cache:
            Whether to cache the key/value representations for future reuse.
        :returns:
            Layer output and the key/value cache.

            *Shape:* ``(batch_size, seq_len, width)``
        """
        residual = input

        attn_out, cache = self.mha(
            self.attn_input_layer_norm(input),
            attention_mask,
            cache=cache,
            store_cache=store_cache,
            positions=positions,
            use_causal_mask=use_causal_mask,
        )
        attn_out = self.attn_output_dropout(attn_out)

        if self.use_parallel_attention:
            ffn_in = input
        else:
            residual = self.attn_residual_layer_norm(input + attn_out)
            ffn_in = residual

        ffn_out = self.ffn(self.ffn_input_layer_norm(ffn_in))
        ffn_out = self.ffn_output_dropout(ffn_out)

        if self.use_parallel_attention:
            output = self.parallel_attn_dropout(attn_out + ffn_out)
        else:
            output = ffn_out

        return self.ffn_residual_layer_norm(residual + output), cache


class DecoderLayer(_TransformerLayer):
    """
    Transformer decoder layer (`Vaswani et al., 2017`_).

    .. _Vaswani et al., 2017: https://arxiv.org/abs/1706.03762
    """

    def forward(
        self,
        input: Tensor,
        attention_mask: AttentionMask,
        *,
        cache: Optional[KeyValueCache] = None,
        positions: Optional[Tensor] = None,
        store_cache: bool = False,
    ) -> Tuple[Tensor, Optional[KeyValueCache]]:
        """
        Apply the decoder layer to the given piece hidden representations.

        :param input:
            Hidden representations to apply the layer to.

            *Shape:* ``(batch_size, seq_len, width)``
        :param attention_mask:
            Attention mask. Sequence elements for which the
            corresponding mask element is set to ``False`` are ignored
            during attention calculation.
        :param cache:
            Key/value cache to avoid recomputing
            key/value representations for tokens that were previously seen.
        :param positions:
            Input positions. Positions are needed to look up rotary embeddings.
            Normally, these positions are calculated automatically. But if the
            positions deviate for some reason, they can be provided through this argument.
        :param store_cache:
            Whether to cache the key/value representations for future reuse.
        :returns:
            Layer output and the key/value cache.

            *Shape:* ``(batch_size, seq_len, width)``
        """
        return super()._forward(
            input,
            attention_mask=attention_mask,
            cache=cache,
            use_causal_mask=True,
            positions=positions,
            store_cache=store_cache,
        )


class EncoderLayer(_TransformerLayer):
    """
    Transformer encoder layer (`Vaswani et al., 2017`_).

    .. _Vaswani et al., 2017: https://arxiv.org/abs/1706.03762
    """

    def forward(
        self,
        input: Tensor,
        attention_mask: AttentionMask,
    ) -> Tuple[Tensor, Optional[KeyValueCache]]:
        """
        Apply the encoder layer to the given piece hidden representations.

        :param input:
            Hidden representations to apply the layer to.

            *Shape:* ``(batch_size, seq_len, width)``
        :param attention_mask:
            Attention mask. Sequence elements for which the
            corresponding mask element is set to ``False`` are ignored
            during attention calculation.
        :returns:
            Layer output and the key/value cache.

            *Shape:* ``(batch_size, seq_len, width)``
        """
        return super()._forward(
            input, attention_mask=attention_mask, use_causal_mask=False
        )



import pytest

from curated_transformers.tokenizers.legacy.bert_tokenizer import BERTTokenizer

from ...compat import has_hf_transformers
from ..util import compare_tokenizer_outputs_with_hf_tokenizer


@pytest.mark.skipif(not has_hf_transformers, reason="requires huggingface transformers")
@pytest.mark.parametrize(
    "model_name",
    [
        "jonfd/electra-small-nordic",
        "Maltehb/aelaectra-danish-electra-small-cased",
        "google/electra-small-discriminator",
    ],
)
def test_from_hf_hub_equals_hf_tokenizer(model_name: str, sample_texts):
    compare_tokenizer_outputs_with_hf_tokenizer(sample_texts, model_name, BERTTokenizer)



from functools import partial
from typing import Any, Dict, Mapping, Optional, Tuple, Type, TypeVar

import torch
from torch import Tensor
from torch.nn import Dropout, LayerNorm, ModuleList

from ...layers.attention import (
    AttentionHeads,
    AttentionLinearBiases,
    QkvMode,
    QkvSplitGroupedByKVHeads,
    ScaledDotProductAttention,
    SelfAttention,
)
from ...layers.embeddings import QueryKeyRotaryEmbeddings
from ...layers.feedforward import PointwiseFeedForward
from ...layers.transformer import (
    DecoderLayer,
    EmbeddingDropouts,
    EmbeddingLayerNorms,
    TransformerDropouts,
    TransformerEmbeddings,
    TransformerLayerNorms,
)
from ..hf_hub import FromHF
from ..hf_hub.conversion import state_dict_from_hf, state_dict_to_hf
from ..transformer import TransformerDecoder
from ._hf import DECODER_HF_PARAM_KEY_TRANSFORMS, _config_from_hf, _config_to_hf
from .config import FalconConfig
from .layer import OldFalconDecoderLayer

# Only provided as typing.Self in Python 3.11+.
Self = TypeVar("Self", bound="FalconDecoder")


class FalconDecoder(TransformerDecoder[FalconConfig], FromHF[FalconConfig]):
    """
    Falcon (`Penedo et al., 2019`_) decoder.

    .. _Penedo et al., 2019: https://arxiv.org/abs/2306.01116
    """

    def __init__(
        self, config: FalconConfig, *, device: Optional[torch.device] = None
    ) -> None:
        """
        Construct a Falcon decoder.

        :param config:
            Decoder configuration.
        :param device:
            Device to which the module is to be moved.
        :returns:
            The decoder.
        """
        super().__init__(config)

        self.embeddings = TransformerEmbeddings(
            dropouts=EmbeddingDropouts(
                embed_output_dropout=Dropout(config.embedding.dropout_prob)
            ),
            embedding_width=config.embedding.embedding_width,
            hidden_width=config.layer.feedforward.hidden_width,
            layer_norms=EmbeddingLayerNorms(),
            n_pieces=config.embedding.n_pieces,
            n_positions=None,
            n_types=None,
            device=device,
        )

        if config.new_decoder_architecture:
            decoder_layer = partial(
                self._create_new_decoder_architecture_layer, config, device
            )
        else:
            decoder_layer = partial(
                self._create_old_decoder_architecture_layer, config, device
            )

        self.layers = ModuleList(
            [decoder_layer() for _ in range(config.layer.n_hidden_layers)]
        )

        self.output_layer_norm = LayerNorm(
            config.layer.feedforward.hidden_width,
            config.layer.layer_norm_eps,
            device=device,
        )

    @classmethod
    def is_supported(cls: Type[Self], config: Dict[str, Any]) -> bool:
        return config.get("model_type") in ("falcon", "RefinedWeb", "RefinedWebModel")

    @classmethod
    def state_dict_from_hf(
        cls: Type[Self], params: Mapping[str, Tensor]
    ) -> Mapping[str, Tensor]:
        return state_dict_from_hf(params, DECODER_HF_PARAM_KEY_TRANSFORMS)

    @classmethod
    def state_dict_to_hf(
        cls: Type[Self], params: Mapping[str, Tensor]
    ) -> Mapping[str, Tensor]:
        return state_dict_to_hf(params, DECODER_HF_PARAM_KEY_TRANSFORMS)

    @classmethod
    def config_from_hf(cls, hf_config: Mapping[str, Any]) -> FalconConfig:
        return _config_from_hf(hf_config)

    @classmethod
    def config_to_hf(cls, curated_config: FalconConfig) -> Mapping[str, Any]:
        return _config_to_hf(cls, curated_config)

    @classmethod
    def from_hf_config(
        cls: Type[Self],
        *,
        hf_config: Any,
        device: Optional[torch.device] = None,
    ) -> Self:
        config = cls.config_from_hf(hf_config)
        return cls(config, device=device)

    def _create_old_decoder_architecture_layer(
        self, config: FalconConfig, device: Optional[torch.device]
    ):
        return OldFalconDecoderLayer(config.layer, device=device)

    def _create_new_decoder_architecture_layer(
        self, config: FalconConfig, device: Optional[torch.device]
    ):
        if config.layer.attention.rotary_embeddings is None:
            raise ValueError(
                "Falcon attention config does not contain rotary embedding parameters"
            )

        hidden_width = config.layer.feedforward.hidden_width
        layer_norm = partial(
            LayerNorm,
            hidden_width,
            config.layer.layer_norm_eps,
            device=device,
        )
        n_attention_heads = config.layer.attention.n_query_heads
        attention_biases = (
            AttentionLinearBiases(
                n_attention_heads=config.layer.attention.n_query_heads,
                is_causal=True,
                is_inverted=True,
            )
            if config.layer.attention.use_alibi
            else None
        )
        # Rotary embeddings are disabled when using ALiBi.
        rotary_embeds = (
            QueryKeyRotaryEmbeddings(
                fraction=config.layer.attention.rotary_embeddings.rotary_fraction,
                base=config.layer.attention.rotary_embeddings.rotary_base,
                head_width=hidden_width // n_attention_heads,
            )
            if not config.layer.attention.use_alibi
            else None
        )
        return DecoderLayer(
            attention_layer=SelfAttention(
                attention_heads=AttentionHeads.key_value_broadcast(
                    n_query_heads=n_attention_heads,
                    n_key_value_heads=config.layer.attention.n_key_value_heads,
                    qkv_split=QkvSplitGroupedByKVHeads(),
                ),
                attention_scorer=ScaledDotProductAttention(
                    dropout_prob=config.layer.attention.dropout_prob,
                    linear_biases=attention_biases,
                ),
                hidden_width=hidden_width,
                qkv_mode=QkvMode.MERGED_SPLIT_AFTER,
                rotary_embeds=rotary_embeds,
                use_bias=config.layer.attention.use_bias,
                device=device,
            ),
            feed_forward_layer=PointwiseFeedForward(
                activation=config.layer.feedforward.activation.module(),
                hidden_width=hidden_width,
                intermediate_width=config.layer.feedforward.intermediate_width,
                use_bias=config.layer.feedforward.use_bias,
                use_gate=config.layer.feedforward.use_gate,
                device=device,
            ),
            dropouts=TransformerDropouts.parallel_attention_dropout(
                config.layer.dropout_prob
            ),
            layer_norms=TransformerLayerNorms(
                attn_input_layer_norm=layer_norm(),
                ffn_input_layer_norm=layer_norm(),
            ),
            # The new decoder uses parallel attention unconditionally.
            use_parallel_attention=True,
        )



from typing import Dict, Iterable, List, Optional, Tuple, TypeVar

from curated_tokenizers import ByteBPEProcessor

from ..chunks import MergedInputChunks, MergedSpecialPieceChunk
from ..tokenizer import PiecesWithIds
from .legacy_tokenizer import LegacyTokenizer

# Only provided as typing.Self in Python 3.11+.
Self = TypeVar("Self", bound="ByteBPETokenizer")


class ByteBPETokenizer(LegacyTokenizer):
    """
    Piece tokenizer using byte-level byte pair encoding
    (`Gage, 1994`_, `Sennrich et al., 2016`_).

    .. _Gage, 1994: https://dl.acm.org/doi/10.5555/177910.177914
    .. _Sennrich et al., 2016: https://arxiv.org/abs/1508.07909
    """

    def __init__(
        self,
        *,
        vocab: Dict[str, int],
        merges: List[Tuple[str, str]],
        special_pieces: Optional[Dict[str, int]] = None,
    ):
        """
        Construct a byte BPE tokenizer.

        :param vocab:
            The word piece vocabulary.
        :param merges:
            Merges.
        :param special_pieces:
            Special pieces.
        """
        self.special_piece_to_id = {} if special_pieces is None else special_pieces
        self.id_to_special_piece = {v: k for k, v in self.special_piece_to_id.items()}
        vocab.update(self.special_piece_to_id)
        self.processor = ByteBPEProcessor(vocab, merges)

    def piece_to_id(self, piece: str) -> Optional[int]:
        return self.processor.token_to_id(piece)

    def _decode(
        self, input: Iterable[Iterable[int]], skip_special_pieces: bool
    ) -> List[str]:
        input = [
            [
                id
                for id in ids
                if not skip_special_pieces or id not in self.id_to_special_piece
            ]
            for ids in input
        ]
        return [self.processor.decode_from_ids(ids) for ids in input]

    def _encode(self, input: Iterable[MergedInputChunks]) -> PiecesWithIds:
        ids = []
        pieces = []

        for seq in input:
            seq_ids = []
            seq_pieces = []

            for chunk in seq:
                if isinstance(chunk, MergedSpecialPieceChunk):
                    piece_id = self.processor.piece_to_id(chunk.piece)
                    if piece_id is None:
                        raise ValueError(f"Unknown special piece: {chunk.piece}")
                    seq_ids.append(piece_id)
                    seq_pieces.append(chunk.piece)
                else:
                    for idx, token in enumerate(chunk.text.split(" ")):
                        if idx != 0:
                            token = " " + token
                        token_ids, token_pieces = self.processor.encode(token)
                        seq_ids.extend(token_ids)
                        seq_pieces.extend(token_pieces)

            ids.append(seq_ids)
            pieces.append(seq_pieces)

        return PiecesWithIds(ids=ids, pieces=pieces)
